#include <cstdint>
#include <memory>
#include <string_view>
#include <algorithm>

using Pixel = std::uint32_t;

class Surface
{
public:
    Surface() noexcept :
        width_ { 0 },
        height_{ 0 },
        buffer_{ nullptr }
    { }
    Surface(int32_t const width, int32_t const height)
    {
        if (width <= 0 || height <= 0)
        {
            width_ = height_ = 0;
            buffer_ = nullptr;
            return;
        }
        size_t const byte_count = width * height * sizeof(Pixel);
        width_  = width;
        height_ = height;
        buffer_ = static_cast<Pixel*>(_aligned_malloc(byte_count, 16));
        if (buffer_ == nullptr)
        {
            throw std::bad_alloc{};
        }
        std::fill_n(buffer_, width_ * height_, 0);
    }
    Surface(Surface const& rhs) :
        Surface{ rhs.width_, rhs.height_ }
    {
        if (buffer_ != nullptr)
        {
            std::copy_n(rhs.buffer_, width_ * height_, buffer_);
        }
    }
    Surface& operator=(Surface const& rhs)
    {
        auto copy = rhs;
        swap(copy);
        return *this;
    }
    Surface(Surface&& rhs) noexcept :
        Surface{}
    {
        swap(rhs);
    }
    Surface& operator=(Surface&& rhs) noexcept
    {
        auto move = std::move(rhs);
        swap(move);
        return *this;
    }
    ~Surface() noexcept
    {
        if (buffer_ != nullptr)
        {
            _aligned_free(buffer_);
        }
    }

    void swap(Surface& rhs) noexcept
    {
        std::swap(buffer_, rhs.buffer_);
        std::swap(width_ , rhs.width_ );
        std::swap(height_, rhs.height_);
    }

private:
    int32_t width_;
    int32_t height_;
    Pixel*  buffer_;
};

inline void swap(Surface& lhs, Surface& rhs) noexcept
{
    lhs.swap(rhs);
}

struct GameObject { Surface image; };
static void example_copy()
{
    Surface image{ 128, 256 };
    GameObject player{ image };
}

struct GameObjectUptr { std::unique_ptr<Surface> image; };
static void example_uptr()
{
    auto image = std::make_unique<Surface>(128, 256);
    GameObjectUptr player{ std::move(image) };
}

static void example_move()
{
    Surface image{ 128, 256 };
    GameObject player{ std::move(image) };
}

static void example_move_assign()
{
    GameObject player;
    Surface image{ 128, 256 };
    player.image = std::move(image);
}
