#include <vector>
#include <memory>
#include <type_traits>
#include <random>

namespace
{

class Component
{
public:
    virtual void update(float delta_seconds) = 0;
    virtual ~Component() noexcept = default;
};

struct Particle
{
    float x, y;
    float vx, vy;
};
class ParticleSystem : public Component
{
    void update(float const delta_seconds) override
    {
        for (auto& particle : particles)
        {
            particle.x += particle.vx * delta_seconds;
            particle.y += particle.vy * delta_seconds;
        }
    }
public:
    std::vector<Particle> particles;
};

class GameObject
{
public:
    //template <typename T>
    //T& create_component()
    //{
    //    components_.push_back(std::make_unique<T>());
    //    return static_cast<T&>(*components_.back());
    //}

    //template <typename T, typename P>
    //T& create_component(P param)
    //{
    //    components_.push_back(std::make_unique<T>(std::move(param)));
    //    return static_cast<T&>(*components_.back());
    //}

    //template <typename T, typename P>
    //T& create_component(P const& param)
    //{
    //    components_.push_back(std::make_unique<T>(param));
    //    return static_cast<T&>(*components_.back());
    //}

    //template <typename T, typename P>
    //T& create_component(P&& param)
    //{
    //    components_.push_back(std::make_unique<T>(std::forward<P>(param)));
    //    return static_cast<T&>(*components_.back());
    //}

    template <typename T, typename... P>
    T& create_component(P&&... params)
    {
        static_assert(std::is_base_of_v<Component, T>, "T must be a Component");
        components_.push_back(std::make_unique<T>(std::forward<P>(params)...));
        return static_cast<T&>(*components_.back());
    }

private:
    std::vector<std::unique_ptr<Component>> components_;
};

void create_fire_particles(ParticleSystem& ps)
{
    ps.particles.clear();
    std::default_random_engine rnd;
    for (int i = 0; i < 100; i++)
    {
        ps.particles.push_back(
        {
            /*  x */std::uniform_real_distribution<float>{ -0.2f, 0.2f }(rnd),
            /*  y */std::uniform_real_distribution<float>{ -0.2f, 0.2f }(rnd),
            /* vx */std::uniform_real_distribution<float>{ -0.2f, 0.2f }(rnd),
            /* vy */std::uniform_real_distribution<float>{  2.0f, 8.0f }(rnd),
        });
    }
}

void example_basic()
{
    GameObject fire;
    auto& particle_system = fire.create_component<ParticleSystem>();
    create_fire_particles(particle_system);
}

void example_multiple()
{
    ParticleSystem ps;
    create_fire_particles(ps);

    GameObject fire;
    fire.create_component<ParticleSystem>(ps);

    GameObject another_fire;
    another_fire.create_component<ParticleSystem>(std::move(ps));
}

//void example_error()
//{
//    GameObject fire;
//    fire.create_component<int>(5);
//}

} // namespace
