#include "exercise.hpp"
#include "test_allocator.hpp"
#include <utility>  // std::move
#include <limits>   // std::numeric_limits<unsigned char>
#include <string.h> // memcpy/memmove
#include <assert.h> // assert macro

namespace
{

constexpr auto bits_per_byte = std::numeric_limits<unsigned char>::digits;

// Calculates number of bytes required to store the number
// of bits. Mathematically: ceil(bits / bits_per_byte)
constexpr size_t bits_to_bytes(size_t const bits)
{
    return (bits_per_byte - 1 + bits) / bits_per_byte;
}

struct offset
{
    size_t byte_index;
    unsigned char bitmask;
};
constexpr offset calculate_offset(size_t const index)
{
    return
    {
        index / bits_per_byte,
        static_cast<unsigned char>(1 << (index % bits_per_byte))
    };
}

} // namespace

// TODO! Define copy/move ctor/assignment here!

BitArray::BitArray(size_t const size, bool const default_value) :
    size_{ size },
    bits_{ size != 0 ? TestAllocator::allocate(bits_to_bytes(size)) : nullptr }
{
    if (size != 0)
    {
        memset(bits_, default_value ? 0xff : 0, bits_to_bytes(size));
    }
}

BitArray::~BitArray() noexcept
{
    if (bits_ != nullptr)
    {
        TestAllocator::deallocate(bits_);
    }
}

void BitArray::swap(BitArray& rhs) noexcept
{
    std::swap(size_, rhs.size_);
    std::swap(bits_, rhs.bits_);
}

bool BitArray::get(size_t const index) const
{
    assert(index < size_);
    // C++17 syntax: Structured Bindings
    // It is roughly equivalent to:
    // auto&& res = calculate_offset(index);
    // auto byte_index = res.byte_index;
    // auto bitmask    = res.bitmask;
    auto[byte_index, bitmask] = calculate_offset(index);
    return (bits_[byte_index] & bitmask) != 0;
}

void BitArray::set(size_t const index, bool const value) const
{
    assert(index < size_);
    auto[byte_index, bitmask] = calculate_offset(index);
    if (value)
    {
        bits_[byte_index] |= bitmask;
    }
    else
    {
        bits_[byte_index] &= ~bitmask;
    }
}
