#include "test_allocator.hpp"
#include "CppUnitTest.h"
#include <unordered_set>
#include <new>      // std::bad_alloc
#include <malloc.h> // malloc/free
#include <memory>


using Assert = Microsoft::VisualStudio::CppUnitTestFramework::Assert;

namespace
{

bool always_throw_{ false };
size_t pending_allocations_{ 0 };
size_t pending_bytes_      { 0 };
size_t total_allocations_  { 0 };
size_t total_bytes_        { 0 };

using Set = std::unordered_set<void*>;

// Using a unique_ptr here because invoking .clear() or copy
// assigning would occasionally hang the test runner. This is
// likely caused by undefined behavior and the order in which
// tests are ran.
std::unique_ptr<Set> allocations_{ std::make_unique<Set>() };

}

static_assert(sizeof(size_t) <= sizeof(max_align_t), "header size must fit in max alignment unit");

void* tracked_malloc(size_t const size)
{
    // Pretend to be out of memory
    if (always_throw_)
    {
        throw std::bad_alloc{};
    }

    auto const memory = malloc(size + sizeof(max_align_t));
    if (memory == nullptr)
    {
        throw std::bad_alloc{};
    }

    // Track stats
    ++pending_allocations_;
    pending_bytes_ += size;
    ++total_allocations_;
    total_bytes_ += size;

    // Write header
    *static_cast<size_t*>(memory) = size;

    auto const data = static_cast<void*>(static_cast<max_align_t*>(memory) + 1);
#pragma warning(push)
#pragma warning(disable: 4101) // unreferenced local variable
    // C++17 Structured Binding
    auto[_, success] = allocations_->insert(data);
    Assert::IsTrue(success, L"allocated the same pointer twice, manually freed?");
#pragma warning(pop)
    return data;
}

void tracked_free(void* const data) noexcept
{
    Assert::IsNotNull(data, L"deallocate called on nullptr");
    Assert::IsTrue(pending_allocations_ > 0, L"deallocate called without pending allocations");
    auto const allocation = allocations_->find(data);
    if (allocation == allocations_->end())
    {
        Assert::Fail(L"deallocate called on pointer that did not come from allocate");
    }
    allocations_->erase(allocation);
    
    auto const memory = static_cast<void*>((static_cast<max_align_t*>(data) - 1));

    // Read header
    auto const size = *static_cast<size_t*>(memory);
    Assert::IsTrue(size <= pending_bytes_, L"deallocate called on an object larger than pending bytes");

    // Track stats
    --pending_allocations_;
    pending_bytes_ -= size;

    free(memory);
}

unsigned char* TestAllocator::allocate(size_t const size)
{
    return static_cast<unsigned char*>(tracked_malloc(size));
}

void TestAllocator::deallocate(unsigned char* const data)
{
    tracked_free(data);
}

void TestAllocator::reset() noexcept
{
    auto const has_leak =
        pending_allocations_ != 0 ||
        pending_bytes_       != 0 ||
        allocations_->size()  != 0;

    pending_allocations_ = pending_bytes_ = 0;
    total_allocations_   = total_bytes_   = 0;
    always_throw_ = false;
    allocations_ = std::make_unique<Set>();

    Assert::IsFalse(has_leak, L"test leaked memory");
}

void TestAllocator::always_throw(bool const value) noexcept
{
    always_throw_ = value;
}

size_t TestAllocator::pending_allocations() noexcept { return pending_allocations_; }
size_t TestAllocator::pending_bytes      () noexcept { return pending_bytes_;       }
size_t TestAllocator::total_allocations  () noexcept { return total_allocations_;   }
size_t TestAllocator::total_bytes        () noexcept { return total_bytes_;         }
