#include "CppUnitTest.h"
#include "test_allocator.hpp"
#include "exercise.hpp"

using namespace Microsoft::VisualStudio::CppUnitTestFramework;

#define ASSERT_ALLOC_STATE(PENDING_ALLOC, PENDING_BYTES, TOTAL_ALLOC, TOTAL_BYTES) ( \
    Assert::AreEqual<size_t>((PENDING_ALLOC), TestAllocator::pending_allocations()), \
    Assert::AreEqual<size_t>((PENDING_BYTES), TestAllocator::pending_bytes()), \
    Assert::AreEqual<size_t>((TOTAL_ALLOC),   TestAllocator::total_allocations()), \
    Assert::AreEqual<size_t>((TOTAL_BYTES),   TestAllocator::total_bytes()), \
    0 \
)

namespace exercise1_move_operations
{		

TEST_CLASS(Exercise1Tests)
{
public:
    TEST_METHOD_INITIALIZE(ResetAllocatorStart)
    {
        TestAllocator::reset();
    }
    TEST_METHOD_CLEANUP(ResetAllocatorEnd)
    {
        TestAllocator::reset();
    }

    TEST_METHOD(DefaultCtor_NoAllocation)
    {
        BitArray a;
        ASSERT_ALLOC_STATE(0, 0, 0, 0);
    }

    TEST_METHOD(ParamCtor_Zero_NoAllocation)
    {
        BitArray a{ 0 };
        Assert::AreEqual<size_t>(0, a.size());
        ASSERT_ALLOC_STATE(0, 0, 0, 0);
    }

    TEST_METHOD(ParamCtor_AllocatesCorrectly)
    {
        {
            BitArray a{ 1 };
            Assert::AreEqual<size_t>(1, a.size());
            ASSERT_ALLOC_STATE(1, 1, 1, 1);
        }
        {
            BitArray a{ 9 };
            Assert::AreEqual<size_t>(9, a.size());
            ASSERT_ALLOC_STATE(1, 2, 2, 3);
        }
        {
            BitArray a{ 8 };
            Assert::AreEqual<size_t>(8, a.size());
            ASSERT_ALLOC_STATE(1, 1, 3, 4);
            BitArray b{ 16 };
            Assert::AreEqual<size_t>(16, b.size());
            ASSERT_ALLOC_STATE(2, 3, 4, 6);
        }
        ASSERT_ALLOC_STATE(0, 0, 4, 6);
    }

    TEST_METHOD(Copy_IsCopyable)
    {
        Assert::IsTrue(std::is_copy_constructible_v<BitArray>);
        Assert::IsTrue(std::is_copy_assignable_v<BitArray>);
    }

    TEST_METHOD(Copy_CopiesAllocation)
    {
        BitArray a;
        ASSERT_ALLOC_STATE(0, 0, 0, 0);
        {
            BitArray b{ 1 };
            ASSERT_ALLOC_STATE(1, 1, 1, 1);
            BitArray c{ b };
            Assert::AreEqual<size_t>(1, c.size());
            ASSERT_ALLOC_STATE(2, 2, 2, 2);
            a = c;
            Assert::AreEqual<size_t>(1, a.size());
            ASSERT_ALLOC_STATE(3, 3, 3, 3);
        }
        ASSERT_ALLOC_STATE(1, 1, 3, 3);
    }

    TEST_METHOD(Copy_NoAllocation_WhenSizeIsSimilar)
    {
        BitArray a{ 9 };
        BitArray b{ 16 };
        ASSERT_ALLOC_STATE(2, 4, 2, 4);
        a = b;
        ASSERT_ALLOC_STATE(2, 4, 2, 4);
    }

    TEST_METHOD(Copy_ExceptionSafety)
    {
        {
            BitArray a{ 9 };
            BitArray b;
            TestAllocator::always_throw();
            Assert::ExpectException<std::bad_alloc>([&]()
            {
                b = a;
            });
            Assert::AreEqual<size_t>(9, a.size());
            Assert::AreEqual<size_t>(0, b.size());
            ASSERT_ALLOC_STATE(1, 2, 1, 2);
        }
        ASSERT_ALLOC_STATE(0, 0, 1, 2);
    }

    TEST_METHOD(ParamCtor_Value_Get)
    {
        BitArray a{ 9 };
        BitArray b{ 9, true };
        ASSERT_ALLOC_STATE(2, 4, 2, 4);
        for (size_t i = 0; i < 9; i++)
        {
            Assert::AreEqual(false, a.get(i));
            Assert::AreEqual(true,  b.get(i));
        }
    }

    TEST_METHOD(Value_Set_Get)
    {
        BitArray a{ 13 };
        for (size_t i = 2; i < 13; i += 3)
        {
            a.set(i, true);
        }
        auto b = a;
        for (size_t i = 0; i < 13; ++i)
        {
            auto const expected = i % 3 == 2;
            Assert::AreEqual(expected, a.get(i));
            Assert::AreEqual(expected, b.get(i));
        }
    }

    TEST_METHOD(Copy_Self_Assign)
    {
        BitArray a{ 13 };
        a.set(4, true);
        a.set(9, true);
        a = a;
        ASSERT_ALLOC_STATE(1, 2, 1, 2);
        for (size_t i = 0; i < 13; i++)
        {
            auto const expected = i == 4 || i == 9;
            Assert::AreEqual(expected, a.get(i));
        }
    }

    TEST_METHOD(Move_IsNoExcept)
    {
        Assert::IsTrue(std::is_nothrow_move_constructible_v<BitArray>);
        Assert::IsTrue(std::is_nothrow_move_assignable_v<BitArray>);
    }

    TEST_METHOD(Move_Default_NoAllocation)
    {
        BitArray a;
        BitArray b = std::move(a);
        Assert::AreEqual<size_t>(0, a.size());
        Assert::AreEqual<size_t>(0, b.size());
        ASSERT_ALLOC_STATE(0, 0, 0, 0);
    }

    TEST_METHOD(Move_NoCopy)
    {
        BitArray a{ 9 };
        BitArray b = std::move(a);
        Assert::AreEqual<size_t>(0, a.size());
        Assert::AreEqual<size_t>(9, b.size());
        ASSERT_ALLOC_STATE(1, 2, 1, 2);
    }

    TEST_METHOD(Move_Assign)
    {
        BitArray a;
        {
            BitArray b{ 9 };
            ASSERT_ALLOC_STATE(1, 2, 1, 2);
            a = std::move(b);
            Assert::AreEqual<size_t>(0, b.size());
            ASSERT_ALLOC_STATE(1, 2, 1, 2);
        }
        Assert::AreEqual<size_t>(9, a.size());
        ASSERT_ALLOC_STATE(1, 2, 1, 2);
    }

    TEST_METHOD(Move_Self_Assign)
    {
        BitArray a{ 13 };
        a.set(4, true);
        a.set(9, true);
        a = std::move(a);
        ASSERT_ALLOC_STATE(1, 2, 1, 2);
        for (size_t i = 0; i < 13; i++)
        {
            auto const expected = i == 4 || i == 9;
            Assert::AreEqual(expected, a.get(i));
        }
    }

    TEST_METHOD(Move_NoAllocationInvoked)
    {
        BitArray a{ 9 };
        BitArray b;
        TestAllocator::always_throw();
        b = std::move(a);
        Assert::AreEqual<size_t>(0, a.size());
        Assert::AreEqual<size_t>(9, b.size());
        ASSERT_ALLOC_STATE(1, 2, 1, 2);
    }
};

} // namespace exercise1_move_operations

#undef ASSERT_ALLOC_STATE
