#include "CppUnitTest.h"
#include "make_sstream.hpp"
#include <string>
#include <iomanip>

using namespace Microsoft::VisualStudio::CppUnitTestFramework;
using namespace std::string_literals;

namespace
{

size_t copy_count = 0;
size_t move_count = 0;

struct WrappedString
{
    WrappedString(WrappedString const& rhs) : str{ rhs.str }
    {
        ++copy_count;
    }
    WrappedString& operator=(WrappedString const& rhs)
    {
        ++copy_count;
        str = rhs.str;
        return *this;
    }
    WrappedString(WrappedString&& rhs) noexcept : str{ std::move(rhs.str) }
    {
        ++move_count;
    }
    WrappedString& operator=(WrappedString&& rhs) noexcept
    {
        ++move_count;
        str = std::move(rhs.str);
        return *this;
    }

    explicit WrappedString(std::string&& s) : str{ s } {}
    std::string str;
};

std::ostream& operator<<(std::ostream& os, WrappedString ws)
{
    return os << ws.str;
}

} // namespace

namespace exercise2_make_sstream
{

TEST_CLASS(Exercise2Tests)
{
public:

    TEST_METHOD_INITIALIZE(reset_counters)
    {
        copy_count = move_count = 0;
    }

    TEST_METHOD(NoParameters)
    {
        auto ss = make_sstream();
        Assert::AreEqual(""s, ss.str());
    }

    TEST_METHOD(BasicSstream)
    {
        auto ss = make_sstream("test");
        Assert::AreEqual("test"s, ss.str());
    }

    TEST_METHOD(MultipleParameters)
    {
        auto ss = make_sstream(42, "test", 5, '\n');
        Assert::AreEqual("42test5\n"s, ss.str());
    }

    TEST_METHOD(IoManips)
    {
        auto ss = make_sstream(std::setw(4), std::hex, 42, std::setw(8), std::setfill('-'), std::left, "hello");
        Assert::AreEqual("  2ahello---"s, ss.str());
    }

    TEST_METHOD(SingleCopy)
    {
        WrappedString ws{ "test"s };
        auto ss = make_sstream(ws);
        Assert::AreEqual("test"s, ss.str());
        Assert::AreEqual<size_t>(1, copy_count);
        Assert::AreEqual<size_t>(0, move_count);
    }

    TEST_METHOD(SingleMove)
    {
        auto ss = make_sstream(WrappedString{ "test"s });
        Assert::AreEqual("test"s, ss.str());
        Assert::AreEqual<size_t>(0, copy_count);
        Assert::AreEqual<size_t>(1, move_count);
    }

};

} // namespace exercise2_make_sstream
